package ch.hepia.numeric;

import java.util.List;
import java.util.ArrayList;
import java.util.function.DoubleFunction;
import java.util.function.Function;

final public class Vector {
    private List<Double> values;

    private Vector() {
        this.values = new ArrayList<>();
    }
    private Vector(Double... elements) {
        List<Double> vec = new ArrayList<>();
        for (double elem : elements){vec.add(elem);}
        this.values = vec;
    }
    private Vector(List<Double> elements) {
        this.values = elements;
    }

    public Transposed t() {
        return new Transposed(this);
    }
    public List<Double> getValues(){
        return this.values;
    }

    public int len() {
        return this.getValues().size(); 
    }

    public double get(int i) {
        return this.getValues().get(i);
    }
    void set(int i, double value) {
        this.values.set(i, value);
    }
    

    public Vector add(Vector that) {
        List<Double> newVec = new ArrayList<>();
        for(int i=0; i < this.len();i++){
            newVec.add(this.get(i)+that.get(i));    
        }
        return new Vector(newVec);
    }

    public Vector mul(double m) {
        List<Double> newVec = new ArrayList<>();
        for (int i=0; i<this.len();i++){
            newVec.add(this.get(i)*m);
        }
        return new Vector(newVec);
    }

    public Vector sub(Vector that) {
        return this.add(that.mul(-1.0));
    }

    public double norm() {
        double sum =  0.0;     
        for (int i=0; i<this.len();i++){
            sum += Math.pow(this.get(i),2.0);
        }
        return Math.sqrt(sum);
    }

    public Vector sliceFrom(int i) {
        List<Double> newVec = new ArrayList<>();
        for (int j=0; j < this.len();j++){
            if (j >= i){
                newVec.add(this.get(j));
            }
        }
        return new Vector(newVec);
    }

    public Vector sliceTo(int i) {
        List<Double> newVec = new ArrayList<>();
        for (int j=0; j < this.len();j++){
            if (j < i){
                newVec.add(this.get(j));
            }
        }
        return new Vector(newVec);
    }

    public Vector slice(int from, int to) {
        Vector newVec = this;
        newVec = newVec.sliceFrom(from);
        newVec = newVec.sliceTo(to-from);
        return newVec;
    }

    public Vector removed(int i) {
        List<Double> newVec = this.getValues();
        newVec.remove(i);
        return new Vector(newVec);
    }

    public Vector concat(Vector that) {
        List<Double> newVec = this.getValues();
        newVec.addAll(that.getValues());
        return new Vector(newVec);
    }

    public Vector map(DoubleFunction<Double> f) {
        List<Double> accVec = this.getValues();
        List<Double> newVec = new ArrayList<>();
    
        for (Double val : accVec){
            newVec.add(f.apply(val));
        }
        return new Vector(newVec);
    }

    public Vector copy() {
        return new Vector(this.getValues());
    }

    void checkVectorLengthOrThrow(Vector that){
        throw new UnsupportedOperationException("This feature isn't implemented yet");
    }

    public static Vector of(Double... elements) {
        return new Vector(elements);
    }

    public static Vector of(List<Double> elements) {
        return new Vector(elements);
    }

    public static Vector empty() {
        return new Vector();
    }

    public static Vector fill(int nb, double value) {
        List<Double> newVec = new ArrayList<>();
        for (int i = 0; i < nb ; i++){
            newVec.add(value);
        }
        return new Vector(newVec);
    }

    public static Vector zeros(int nb) {
        List<Double> newVec = new ArrayList<>();
        for (int i = 0; i < nb ; i++){
            newVec.add(0.0);
        }
        return new Vector(newVec);
    }

    public static Vector ones(int nb) {
        List<Double> newVec = new ArrayList<>();
        for (int i = 0; i < nb ; i++){
            newVec.add(1.0);
        }
        return new Vector(newVec);
    }

    public static Vector linespace(double from, double to, int nb) {
        double diff = to - from;
        double hop = diff / (nb-1);
        List<Double> newVec = new ArrayList<>();
        for (int i = 0; i <= nb -1; i++){
            newVec.add(hop*i);
        }
        return new Vector(newVec);
    }

    public static Vector tabulate(int nb, Function<Integer, Double> f) {
        List<Double> newVec = new ArrayList<>();
        for (int i=0; i < nb; i++){
            newVec.add(f.apply(i));
        }
        return new Vector(newVec);
    }

    public static Vector sum(List<Vector> vs) {
        Vector totalVec = Vector.zeros(vs.get(0).len());
        for (Vector vec : vs){
            totalVec = totalVec.add(vec);
        }
        return totalVec;
    }

    public static double norms(List<Vector> vs) {
        return Vector.sum(vs).norm();
    }
    // FILL NEW VERSION :
    public static SubVec fill(int nbr){
        return new Vector().new SubVec(nbr);
    }
    protected class SubVec{
        private int nbre;
        protected SubVec(int nbre){
            this.nbre = nbre;
        }
        private int getNbre(){
            return this.nbre;
        }
        
        public Vector withValue(Double alpha){
            List<Double> newVec = new ArrayList<>(); 
            for (int i = 0; i < this.getNbre();i++){
                newVec.add(alpha);
            }
            return new Vector(newVec);
        }
    }

    @Override
    public String toString() {
        String vec = "Vector[";
        for (double val: this.getValues()){
            vec = vec.concat(val + ", ");
        }
        vec = vec.substring(0, vec.length()-2).concat("]");
        return vec;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) return true;
        if (obj == null || obj.getClass() != this.getClass()) {
            return false;
        }
        Vector vec = (Vector) obj;
        if (vec.len() != this.len()){
            return false;
        }
        for (int i=0;i<vec.len();i++){
            if (Double.compare(vec.getValues().get(i), this.getValues().get(i))!=0){
                return false;
            }
        }
        return true;
    }
}